From 366d7a99ed92f1aca150d41dfe5b285ca2718896 Mon Sep 17 00:00:00 2001
From: Pavel Baturko <pabftk@gmail.com>
Date: Sun, 20 Jan 2013 22:05:37 +0000
Subject: [PATCH] Decode UCS2 SMS to UTF8

---
 src/host/layer23/src/mobile/gsm411_sms.c |   19 ++++++++++++++++---
 1 files changed, 16 insertions(+), 3 deletions(-)

diff --git a/src/host/layer23/src/mobile/gsm411_sms.c b/src/host/layer23/src/mobile/gsm411_sms.c
index 9de1259..c3ebfdd 100644
--- a/src/host/layer23/src/mobile/gsm411_sms.c
+++ b/src/host/layer23/src/mobile/gsm411_sms.c
@@ -38,6 +38,7 @@
 #include <osmocom/bb/mobile/transaction.h>
 #include <osmocom/bb/mobile/gsm411_sms.h>
 #include <osmocom/gsm/gsm0411_utils.h>
+#include <osmocom/gsm/gsm_utils.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/bb/mobile/vty.h>
 
@@ -184,12 +185,20 @@ static int gsm340_rx_sms_deliver(struct osmocom_ms *ms, struct msgb *msg,
 	char *sms_file;
 	char vty_text[sizeof(gsms->text)], *p;
 	FILE *fp;
+	uint8_t text_in_utf8 = 0;
 
 	/* remove linefeeds and show at VTY */
 	strcpy(vty_text, gsms->text);
 	for (p = vty_text; *p; p++) {
 		if (*p == '\n' || *p == '\r')
 			*p = ' ';
+		else if ((uint8_t)*p > 0x7F) {
+			text_in_utf8 = 1;
+			break;
+		}
+	}
+	if (text_in_utf8) {
+		strcpy(vty_text, "SMS can't be displayed. See log/sms-file.");
 	}
 	vty_notify(ms, NULL);
 	vty_notify(ms, "SMS from %s: '%s'\n", gsms->address, vty_text);
@@ -225,6 +234,7 @@ static int gsm340_rx_tpdu(struct gsm_trans *trans, struct msgb *msg)
 	uint8_t *smsp = msgb_sms(msg);
 	struct gsm_sms *gsms;
 	unsigned int sms_alphabet;
+	uint8_t have_utf8_locale = 0;
 	uint8_t sms_mti, sms_mms;
 	uint8_t oa_len_bytes;
 	uint8_t address_lv[12]; /* according to 03.40 / 9.1.2.5 */
@@ -285,7 +295,11 @@ static int gsm340_rx_tpdu(struct gsm_trans *trans, struct msgb *msg)
 			gsm_7bit_decode(gsms->text, smsp, gsms->user_data_len);
 			break;
 		case DCS_8BIT_DATA:
+			break;
 		case DCS_UCS2:
+			ucs2_to_utf8_decode(gsms->text, sizeof(gsms->text), gsms->user_data, gsms->user_data_len / 2);
+			have_utf8_locale = check_utf8_locale();
+			break;
 		case DCS_NONE:
 			break;
 		}
@@ -297,9 +311,8 @@ static int gsm340_rx_tpdu(struct gsm_trans *trans, struct msgb *msg)
 	     sms_mti, gsms->msg_ref,
 	     gsms->protocol_id, gsms->data_coding_scheme, gsms->address,
 	     gsms->user_data_len,
-			sms_alphabet == DCS_7BIT_DEFAULT ? gsms->text :
-				osmo_hexdump(gsms->user_data,
-						gsms->user_data_len));
+			(sms_alphabet == DCS_7BIT_DEFAULT || (sms_alphabet == DCS_UCS2 && have_utf8_locale)) ? 
+				gsms->text : osmo_hexdump(gsms->user_data, gsms->user_data_len));
 
 	switch (sms_mti) {
 	case GSM340_SMS_DELIVER_SC2MS:
-- 
1.7.2.5

